import os

from nose.tools import raises

from s3tup.connection import Connection
from s3tup.exception import AwsCredentialNotFound

@raises(AwsCredentialNotFound)
def test_connection_init_no_credentials():
    os.environ.pop('AWS_ACCESS_KEY_ID', None)
    os.environ.pop('AWS_SECRET_ACCESS_KEY', None)
    Connection()

def test_connection_init_explicit_credentials():
    os.environ.pop('AWS_ACCESS_KEY_ID', None)
    os.environ.pop('AWS_SECRET_ACCESS_KEY', None)
    c = Connection('explicit_access_key_id', 'explicit_secret_access_key')
    assert c.access_key_id == 'explicit_access_key_id'
    assert c.secret_access_key == 'explicit_secret_access_key'

def test_connection_init_environ_credentials():
    os.environ['AWS_ACCESS_KEY_ID'] = 'environ_access_key_id'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'environ_secret_access_key'
    c = Connection()
    assert c.access_key_id == 'environ_access_key_id'
    assert c.secret_access_key == 'environ_secret_access_key'

def test_connection_init_explicit_supercedes_environ():
    os.environ['AWS_ACCESS_KEY_ID'] = 'environ'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'environ'
    c = Connection('explicit', 'explicit')
    assert c.access_key_id == 'explicit'
    assert c.secret_access_key == 'explicit'

#hey


from mock import MagicMock

class ConnMock(object):
    def __init__(self):
        self.make_request = MakeRequestMock()

# Needed to ignore redundant default kwargs
# That is, MagicMock didn't recognize that 'data=None'
# is the same as 'data' not being present at all.
# MakeRequestMock removes None/default kwargs all together
# from both __call__ and assert_called_once_with.
# It also reads data if data is a file like object.
class MakeRequestMock(MagicMock):
    def __call__(self, *args, **kwargs):
        kwargs = self.clean_kwargs(kwargs)
        super(MakeRequestMock, self).__call__(*args, **kwargs)

    def assert_called_once_with(self, *args, **kwargs):
        kwargs = self.clean_kwargs(kwargs)
        super(MakeRequestMock, self).assert_called_once_with(*args, **kwargs)

    @staticmethod
    def clean_kwargs(kwargs):
        cleaned = {}
        for k,v in kwargs.iteritems():
            if k == 'headers' and v in (None, {}):
                pass
            elif k == 'data' and v in ('', None):
                pass
            elif k == 'data' and is_readable(v):
                cleaned[k] = v.read()
            elif v is not None:
                cleaned[k] = v
        return cleaned

def is_readable(f):
    return callable(getattr(f, "read", None))


KEY_ATTRS = (
    'acl',
    'cache_control',
    'canned_acl',
    'content_disposition',
    'content_encoding',
    'content_type',
    'content_language',
    'encrypted',
    'expires',
    'metadata',
    'reduced_redundancy',
)

KEY_HEADERS = (
    'cache_control',
    'content_disposition',
    'content_encoding',
    'content_type',
    'content_language',
    'expires',
)

BUCKET_ATTRS = (
    'acl',
    'canned_acl',
    'cors',
    'lifecycle',
    'logging',
    'notification',
    'policy',
    'region',
    'tagging',
    'versioning',
    'website',
    'redirects',
)


class S3ResponseError(Exception):
    def __init__(self, code, message):
        msg = "{}: {}".format(code, message)
        super(S3ResponseError, self).__init__(msg)

class ActionConflict(Exception):
    pass

class ConfigLoadError(Exception):
    pass

class ConfigParseError(Exception):
    pass

class AwsCredentialNotFound(Exception):
    pass

class SecretAccessKeyNotFound(AwsCredentialNotFound):
    pass

class AccessKeyIdNotFound(AwsCredentialNotFound):
    pass


from contextlib import contextmanager
import logging
import copy

from s3tup.connection import Connection
from s3tup.bucket import Bucket
from s3tup.key import KeyFactory, KeyConfigurator
from s3tup.rsync import RsyncPlanner, RsyncConfig
from s3tup.exception import ConfigParseError, ConfigLoadError
from s3tup.utils import Matcher

log = logging.getLogger('s3tup.parse')

def load_config(config):
    """Return parseable s3tup configuration from whatever 'config' is.

    Likely the most magic method here, load_config guesses how to convert
    'config' into a standard s3tup config list-of-dicts based on its type.

    Currently supports:

        * config is basestring: load yaml from file at 'config'
        * config is file: load yaml from config.read()
        * config is dict: put into list
        * config is list: pass it on through

    """
    if isinstance(config, basestring):
        try:
            config = open(config, 'r')
        except IOError:
            msg = "File '{}' does not exist".format(config)
            raise ConfigLoadError(msg)

    read = getattr(config, "read", None)
    if callable(read):
        import yaml
        try:
            config = yaml.load(config)
        except yaml.YAMLError as e:
            msg = "Problem parsing yaml:\n"+ e.__str__()
            raise ConfigLoadError(msg)

    if isinstance(config, dict):
        config = [config,]

    return config

# Recursive config definition
#
# [x, ...] = list of 'x's
# {a, b, c} = dict with keys 'a', 'b', 'c'
# {a!} = key 'a' is required
# {*iterable} = unpack iterable into definition
#
# config: [bucket, ...]
# bucket: {bucket!, key_config, rsync, *s3tup.constants.BUCKET_ATTRS}
# key_config: [key_configurator, ...]
# key_configurator: {*s3tup.constants.KEY_ATTRS, *matcher_fields}
# rsync: (src|rsync_object|[rsync_object,])
# rsync_object: {src, dest, delete, *matcher_fields}
# matcher_fields: (patterns, ignore_patterns, regexes, ignore_regexes)

# IMPORTANT:
# Many parse methods directly mutate input by design. This can lead to many
# hard to debug issues, especially when using parsers programtically (rather
# than through the cli). To counteract this, the parse_method decorator is
# applied to each parse method. This decorator basically calls copy.deepcopy
# on the input before passing it on to the decorated function. Performance
# not really an issue here, anything more fine grained would be premature.

def parse_method(f):
    def inner(config):
        return f(copy.deepcopy(config))
    return inner

@contextmanager
def exception_ctx(context):
    """Append 'context' to all exceptions raised in context"""
    try: yield
    except Exception as e:
        msg = "{}: {}".format(context, e.message)
        raise ConfigParseError(msg)

def convert_type_error(e):
    """Return TypeError converted to more readable ConfigParseError"""
    bad_kwarg = e.message[e.message.index("'")+1:-1]
    msg = "Invalid field '{}'".format(bad_kwarg)
    return ConfigParseError(msg)

@parse_method
def parse_config(config):
    """Return a list of fully configured Buckets from 'config'."""
    if not isinstance(config, list):
        raise ConfigParseError('Config must be a list')

    buckets = []
    for bucket_config in config:
        buckets.append(parse_bucket(bucket_config))
    return buckets

@parse_method
def parse_bucket(config):
    """Return a properly configured Bucket from 'config'."""
    if not isinstance(config, dict):
        msg = 'Every bucket config must be a dict'
        raise ConfigParseError(msg)

    try:
        bucket_name = config.pop('bucket')
    except KeyError:
        msg = "Every bucket config must contain field 'bucket'"
        raise ConfigParseError(msg)

    access_key_id = config.pop('access_key_id', None)
    secret_access_key = config.pop('secret_access_key', None)
    conn = Connection(access_key_id, secret_access_key)

    with exception_ctx(bucket_name):
        if 'key_config' in config:
            key_factory = parse_key_config(config.pop('key_config'))
        else:
            key_factory = None

        if 'rsync' in config:
            rsync_planner = parse_rsync(config.pop('rsync'))
        else:
            rsync_planner = None

        # Attempt to create and return a Bucket from the supplied config.
        # Bucket will throw TypeError if it gets unknown kwargs.
        try:
            return Bucket(conn, bucket_name, key_factory, rsync_planner, **config)
        except TypeError as e:
            invalid = e.message[e.message.index("'")+1:-1]
            msg = "Invalid field '{}'".format(invalid)
            raise ConfigParseError(msg)

@parse_method
def parse_rsync(config):
    """Return a list of Rsync objects from 'config'

    This has some magic in that it'll accept a string, dict, or list.
    Done to make rsync definitions less verbose when they don't need to be.

    Examples of valid yaml configs for the rsync field:

        rsync: path/to/some/folder            # string

        rsync:
          src: path/to/some/folder            # dict
          delete: true

        rsync:
          - src: path/to/some/folder          # list
            delete: true
          - src: path/to/some/other/folder
            dest: some/prefix/

    Regardless of the input type, this function returns an RsyncPlanner.

    """
    # Convert all acceptable input into a standard list-of-dicts format
    if isinstance(config, basestring):
        config = [{'src': config},]
    elif isinstance(config, dict):
        config = [config,]

    rsync_planner = RsyncPlanner()
    with exception_ctx('rsync'):
        for r in config:
            rsync_planner.configs.append(parse_rsync_object(r))
    return rsync_planner

@parse_method
def parse_rsync_object(config):
    """Return a properly configured Rsync object from 'config'"""
    matcher, config = extract_matcher(config)
    try:
        return RsyncConfig(matcher=matcher, **config)
    except TypeError as e:
        raise convert_type_error(e)

@parse_method
def parse_key_config(config):
    """Return a properly configured KeyFactory from 'config'"""
    # key_config must be a list
    if not isinstance(config, list):
        raise ConfigParseError('key_config must be a list')

    configurators = []
    with exception_ctx('key_config'):
        for c in config:
            configurators.append(parse_key_configurator(c))
    return KeyFactory(configurators)

@parse_method
def parse_key_configurator(config):
    """Return a properly configured KeyConfigurator from 'config'"""
    matcher, config = extract_matcher(config)
    try:
        return KeyConfigurator(matcher=matcher, **config)
    except TypeError as e:
        raise convert_type_error(e)

@parse_method
def extract_matcher(config):
    """Return (matcher, config) tuple extracted from 'config'.

    Extracts the matcher fields from the passed in config dict, creates a
    Matcher object from them, and then returns the Matcher and the config
    dict minus the matcher fields back in a tuple. If none of the matcher
    fields are present, the matcher returned will be None.

    """
    if not isinstance(config, dict):
        msg = "Attempt to extract matcher from non-dict: '{}'".format(config)
        raise ConfigParseError(msg)

    patterns = config.pop('patterns', None)
    ignore_patterns = config.pop('ignore_patterns', None)
    regexes = config.pop('regexes', None)
    ignore_regexes = config.pop('ignore_regexes', None)

    params = (patterns, ignore_patterns, regexes, ignore_regexes)
    if any(p is not None for p in params):
        matcher = Matcher(*params)
    else:
        matcher = None
    return matcher, config


from base64 import b64encode
from email.utils import formatdate
from contextlib import contextmanager
import os
import logging
import hashlib
import hmac
import urllib

import gevent
from gevent.pool import Pool
from gevent import monkey
monkey.patch_all(thread=False, select=False)

# Make greenlets not print traceback info on exception.
# I imagine this isn't a good thing to do, but pushing a
# traceback to stdout on a cli is a no go.
from gevent.hub import Hub
Hub.print_exception = lambda *args, **kwargs:None

from bs4 import BeautifulSoup
from requests import Session, Request
from requests.structures import CaseInsensitiveDict

from s3tup.exception import S3ResponseError, AccessKeyIdNotFound, \
                            SecretAccessKeyNotFound
import s3tup.utils as utils

log = logging.getLogger('s3tup.connection')

class Connection(object):

    def __init__(self, access_key_id=None, secret_access_key=None,
                 concurrency=5):
        if access_key_id is None:
            try:
                access_key_id = os.environ['AWS_ACCESS_KEY_ID']
            except KeyError:
                raise AccessKeyIdNotFound()
        if secret_access_key is None:
            try:
                secret_access_key = os.environ['AWS_SECRET_ACCESS_KEY']
            except KeyError:
                raise SecretAccessKeyNotFound()

        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key

        self.concurrency = concurrency
        self._joined = False

        self.stats = {'GET':0, 'POST':0, 'PUT':0, 'DELETE':0, 'HEAD':0}

    @property
    def concurrency(self):
        return self._concurrency

    @concurrency.setter
    def concurrency(self, val):
        try:
            self._pool.join()
        except AttributeError:
            pass
        if val > 0:
            self._pool = Pool(val)
        self._concurrency = val

    # Join requires some strange context management because it's
    # possible for joined methods to themselves call join. If
    # these methods then saturate the pool, the joins that they're
    # waiting on will never complete. To counteract this we allow
    # another greenlet into the pool for the duration of the join
    # call *if* the join is within a join already.
    @contextmanager
    def joincontext(self):
        old = self._joined
        self._joined = True
        if old:
            self._pool._semaphore.counter += 1
            yield
            self._pool._semaphore.counter -= 1
        else:
            yield
        self._joined = old

    def join(self, functions):
        if self.concurrency <= 0: # Useful for debugging
            out = []
            for f in functions:
                if hasattr(f, '__iter__'):
                    out.append(f[0](*f[1:]))
                else:
                    out.append(f())
            return out

        with self.joincontext():
            greenlets = []
            for f in functions:
                if hasattr(f, '__iter__'):
                    greenlet = gevent.spawn(*f)
                else:
                    greenlet = gevent.spawn(f)
                self._pool.add(greenlet)
                greenlets.append(greenlet)
            gevent.joinall(greenlets, raise_error=True)
            return [g.get() for g in greenlets]

    # Here be dragons
    def make_request(self, method, bucket, key=None, params=None, data=None,
                     headers=None):
        # Remove params that are set to None
        if isinstance(params, dict):
            for k,v in params.copy().items():
                if v is None:
                    params.pop(k)

        # Construct target url
        url = 'http://{}.s3.amazonaws.com'.format(bucket)
        url += '/{}'.format(key) if key is not None else '/'
        if isinstance(params, dict) and len(params) > 0:
            url += '?{}'.format(urllib.urlencode(params))
        elif isinstance(params, basestring):
            url += '?{}'.format(params)

        # Make headers case insensitive
        if headers is None:
            headers = {}
        headers = CaseInsensitiveDict(headers)

        headers['Host'] = '{}.s3.amazonaws.com'.format(bucket)

        if data is not None:
            try:
                raw_md5 = utils.f_md5(data)
            except:
                m = hashlib.md5()
                m.update(data)
                raw_md5 = m.digest()
            md5 = b64encode(raw_md5)
            headers['Content-MD5'] = md5
        else:
            md5 = ''

        try: content_type = headers['Content-Type']
        except KeyError: content_type = ''

        date = formatdate(timeval=None, localtime=False, usegmt=True)
        headers['x-amz-date'] = date

        # Construct canonicalized amz headers string
        canonicalized_amz_headers = ''
        amz_keys = [k for k in list(headers.keys()) if k.startswith('x-amz-')]
        for k in sorted(amz_keys):
            v = headers[k].strip()
            canonicalized_amz_headers += '{}:{}\n'.format(k.lower(), v)

        # Construct canonicalized resource string
        canonicalized_resource = '/' + bucket
        canonicalized_resource += '/' if key is None else '/{}'.format(key)
        if isinstance(params, basestring):
            canonicalized_resource += '?{}'.format(params)
        elif isinstance(params, dict) and len(params) > 0:
            canonicalized_resource += '?{}'.format(urllib.urlencode(params))

        # Construct string to sign
        string_to_sign = method.upper() + '\n'
        string_to_sign += md5 + '\n'
        string_to_sign += content_type + '\n'
        string_to_sign += '\n' # date is always set through x-amz-date
        string_to_sign += canonicalized_amz_headers + canonicalized_resource

        # Create signature
        h = hmac.new(self.secret_access_key, string_to_sign, hashlib.sha1)
        signature = b64encode(h.digest())

        # Set authorization header
        auth_head = 'AWS {}:{}'.format(self.access_key_id, signature)
        headers['Authorization'] = auth_head

        # Prepare Request
        req = Request(method, url, data=data, headers=headers).prepare()

        # Log request data.
        # Prepare request beforehand so requests-altered headers show.
        # Combine into a single message so we don't have to bother with
        # locking to make lines appear together.
        log_message = '{} {}\n'.format(method, url)
        log_message += 'headers:'
        for k in sorted(req.headers.keys()):
            log_message += '\n {}: {}'.format(k, req.headers[k])
        log.debug(log_message)

        # Send request
        resp = Session().send(req)
        
        # Update stats, log response data.
        self.stats[method.upper()] += 1
        log.debug('response: {} ({} {})'.format(resp.status_code, method, url))
        
        # Handle errors
        if resp.status_code/100 != 2:
            soup = BeautifulSoup(resp.text)
            error = soup.find('error')

            log_message = "S3 replied with non 2xx response code!!!!\n"
            log_message += '  request: {} {}\n'.format(method, url)
            for c in error.children:
                error_name = c.name
                error_message = c.text.encode('unicode_escape')
                log_message +='  {}: {}\n'.format(error_name, error_message)
            log.debug(log_message)

            code = error.find('code').text
            message = error.find('message').text
            raise S3ResponseError(code, message)

        return resp


from binascii import hexlify
import logging
import mimetypes

from bs4 import BeautifulSoup

import s3tup.utils as utils
import s3tup.constants as constants

log = logging.getLogger('s3tup.key')

MULTIPART_CUTOFF = 5242880
MULTIPART_PART_SIZE = 5242880 # Minimum: 5242880

class KeyFactory(object):
    """
    Basically just a container for KeyConfigurators. make_key will create a
    key based on the input parameters and then run each configurator on it
    sequentially, returning a fully configured key.

    """
    def __init__(self, configurators=None):
        self.configurators = configurators or []

    def make_key(self, conn, bucket_name, key_name,):
        """Return a properly configured key"""
        key = Key(conn, bucket_name, key_name)
        return self.configure_key(key)

    def configure_key(self, key):
        for c in self.configurators:
            if c.effects_key(key.name):
                key = c.configure_key(key)
        return key


class KeyConfigurator(object):
    def __init__(self, matcher=None, **kwargs):
        self.matcher = matcher or utils.Matcher()
        for k, v in kwargs.items():
            if k in constants.KEY_ATTRS:
                self.__dict__[k] = v
            else:
                msg = ("__init__() got an unexpected keyword"
                       " argument '{}'".format(k))
                raise TypeError(msg)

    def effects_key(self, key_name):
        """Return whether this configurator effects key_name"""
        return self.matcher.matches(key_name)

    def configure_key(self, key):
        """Return the input key with all configurations applied"""
        for attr in constants.KEY_ATTRS:
            if attr in self.__dict__ and attr != 'metadata':
                key.__dict__[attr] = self.__dict__[attr]
        try: key.metadata.update(self.metadata)
        except AttributeError: pass
        return key

# Both deletion and redirection disregard key configuration,
# so they can be standalone methods.

def delete_key(conn, bucket, key):
    path = key_pretty_path(bucket, key)
    log.info('delete: {}'.format(path))
    conn.make_request('DELETE', bucket, key)

def redirect_key(conn, bucket, key, redirect_url):
    path = key_pretty_path(bucket, key)
    log.info('redirect: {} -> {}'.format(path, redirect_url))
    headers = {'x-amz-website-redirect-location': redirect_url}
    conn.make_request('PUT', bucket, key, headers=headers)

def key_pretty_path(bucket, key):
    return 's3://{}/{}'.format(bucket, key)

class Key(object):
    """
    Encapsulates configuration for a particular s3 key. It has attributes
    (all defined in constants.KEY_ATTRS) that you can set, delete, modify, 
    and then sync to s3 using the sync or upload methods.

    """
    def __init__(self, conn, bucket_name, key_name, **kwargs):
        self.conn = conn
        self.name = key_name
        self.bucket_name = bucket_name

        # Set defaults for required attributes:
        self.reduced_redundancy = kwargs.pop('reduced_redundancy', False)
        self.encrypted = kwargs.pop('encrypted', False)
        self.metadata = kwargs.pop('metadata', {})

        # Add all kwargs passed in that are named in 
        # constants.KEY_ATTRS to this object instance
        for k,v in kwargs.items():
            if k in constants.KEY_ATTRS:
                self.__dict__[k] = v
            else:
                msg = ("__init__() got an unexpected keyword"
                       " argument '{}'".format(k))
                raise TypeError(msg)

    @property
    def pretty_path(self):
        return key_pretty_path(self.bucket_name, self.name)

    def make_request(self, method, params=None, data=None, headers=None):
        """ Convenience method for self.conn.make_request; has bucket and key
            already filled in.
        """
        return self.conn.make_request(
            method,
            self.bucket_name,
            self.name,
            params=params,
            data=data,
            headers=headers
        )

    def get_headers(self):
        """Return the headers associated with this key"""
        headers = {}

        try:
            headers['x-amz-acl'] = self.canned_acl
        except AttributeError:
            pass

        # No need to make separate acl req if acl=None
        try:
            if self.acl is None:
                headers['x-amz-acl'] = 'private'
        except AttributeError:
            pass

        if self.reduced_redundancy:
            headers['x-amz-storage-class'] = 'REDUCED_REDUNDANCY'

        if self.encrypted:
            headers['x-amz-server-side-encryption'] = 'AES256'

        for k, v in self.metadata.items():
            headers['x-amz-meta-' + k] = v

        for k in constants.KEY_HEADERS:
            if k in self.__dict__:
                headers[k.replace('_', '-')] = self.__dict__[k]

        # Guess content-type
        if 'content-type' not in headers:
            content_type_guess = mimetypes.guess_type(self.name)[0]
            if content_type_guess is not None:
                headers['content-type'] = content_type_guess

        return headers

    def delete(self):
        delete_key(self.conn, self.bucket_name, self.name)

    def redirect(self, redirect_url):
        redirect_key(self.conn, self.bucket_name, self.name, redirect_url)

    def sync(self):
        """Sync this object's configuration with its respective key on s3"""
        log.info("sync: {}".format(self.pretty_path))

        headers = self.get_headers()
        headers['x-amz-copy-source'] = '/'+self.bucket_name+'/'+self.name
        headers['x-amz-metadata-directive'] = 'REPLACE'

        # According to S3 docs, copy response can contain error & info
        # even if it returns 200 OK. Need to handle this. However, the
        # docs don't mention what the error would look like...
        self.make_request('PUT', headers=headers)
        self.sync_acl()

    def upload(self, f):
        """Upload file like object to s3 with this key's configuration"""
        try:
            log.info("upload: {} <- {}".format(self.pretty_path, f.name))
        except AttributeError:
            log.info("upload: {}".format(self.pretty_path))

        if utils.f_sizeof(f) <= MULTIPART_CUTOFF:
            self._basic_upload(f)
        else:
            self._multipart_upload(f)
        self.sync_acl()

    def _basic_upload(self, f):
        headers = self.get_headers()
        self.make_request('PUT', headers=headers, data=f)

    def _multipart_upload(self, f):
        upload_id = self._initiate_multipart_upload()

        # Upload individual parts
        upload_reqs = []
        parts = []
        chunks = utils.f_chunk(f, MULTIPART_PART_SIZE)
        for r in range(len(chunks)):
            upload_reqs.append([
                self._multipart_upload_part,
                chunks[r],
                r+1,
                upload_id
            ])
            parts.append((r+1, hexlify(utils.f_md5(chunks[r]))))
        try:
            self.conn.join(upload_reqs)
        except:
            self._abort_multipart_upload(upload_id)
            raise
        
        self._complete_multipart_upload(upload_id, parts)        

    def _initiate_multipart_upload(self):
        headers = self.get_headers()
        resp = self.make_request('POST', 'uploads', headers=headers)
        return BeautifulSoup(resp.text).find('uploadid').text

    def _complete_multipart_upload(self, upload_id, parts):
        data = "<CompleteMultipartUpload>\n"
        for part in parts:
            data += (
                "<Part>\n"
                "  <PartNumber>{}</PartNumber>\n"
                "  <ETag>\"{}\"</ETag>\n"
                "</Part>\n".format(*part)
            )
        data += "</CompleteMultipartUpload>"
        self.make_request('POST', {'uploadId': upload_id}, data=data)

    def _abort_multipart_upload(self, upload_id):
        self.make_request('DELETE', {'uploadId': upload_id})

    def _multipart_upload_part(self, f, part_num, upload_id):
        log.info("upload: {} (part {})".format(self.pretty_path, part_num))
        params = {'partNumber': part_num, 'uploadId': upload_id}
        return self.make_request('PUT', params=params, data=f)

    def sync_acl(self):
        try: acl = self.acl
        except AttributeError: return False
        
        log.info("sync acl: {}".format(self.pretty_path))
        if acl is not None:
            self.make_request('PUT', 'acl', data=acl)


from StringIO import StringIO
from binascii import hexlify
from tempfile import NamedTemporaryFile

import s3tup.utils as utils

def test_matcher_patterns():
    m = utils.Matcher(patterns=['*.py', '*.md'])
    assert m.matches('test.py')
    assert m.matches('test.md')
    assert not m.matches('test.ps')
    assert not m.matches('testpy')

def test_matcher_ignore_patterns_only():
    m = utils.Matcher(ignore_patterns=['*.py',])
    assert not m.matches('test.py')
    assert m.matches('test.md')
    assert m.matches('test.yml')

def test_matcher_regexes():
    m = utils.Matcher(regexes=['\.py$', '\.md$'])
    assert m.matches('test.py')
    assert m.matches('test.md')
    assert not m.matches('test.ps')
    assert not m.matches('testpy')  

def test_matcher_ignore_regexes_only():
    m = utils.Matcher(ignore_regexes=['\.py$'])
    assert not m.matches('test.py')
    assert m.matches('test.md')
    assert m.matches('test.yml')

def test_matcher_ignore_overrides():
    m = utils.Matcher(['*.py'], ['ignore*.py'])
    assert m.matches('test.py')
    assert not m.matches('test.md')
    assert not m.matches('ignore_test.py')

def test_matcher_makes_sets():
    m = utils.Matcher(patterns=['test', 'test'])
    assert isinstance(m.patterns, set)
    assert isinstance(m.ignore_patterns, set)
    assert len(m.patterns) == 1
    assert len(m.ignore_patterns) == 0

def test_matcher_default_matches_everything():
    m = utils.Matcher()
    assert m.matches('test.py')
    assert m.matches('test.md')
    assert m.matches('test.ps')
    assert m.matches('testpy')

def test_f_sizeof():
    s = StringIO('hello')
    s.seek(2, 0)
    assert utils.f_sizeof(s) == 5
    assert s.tell() == 2

def test_f_md5():
    s = StringIO('hello')
    s.seek(2, 0) 
    expected = '5d41402abc4b2a76b9719d911017c592'
    assert hexlify(utils.f_md5(s)) == expected
    assert s.tell() == 2

def test_f_chunk_even():
    f = NamedTemporaryFile()
    f.write('01234567')
    chunks = utils.f_chunk(f, 2)
    assert len(chunks) == 4
    for r in range(4):
        assert utils.f_sizeof(chunks[r]) == 2
        assert chunks[r].read() == '{}{}'.format(r*2, r*2+1)
        chunks[r].close()
        assert chunks[r].closed
    f.close()

def test_f_chunk_not_odd():
    f = NamedTemporaryFile()
    f.write('012345678')
    chunks = utils.f_chunk(f, 2)
    assert len(chunks) == 5
    for r in range(4):
        assert utils.f_sizeof(chunks[r]) == 2
        assert chunks[r].read() == '{}{}'.format(r*2, r*2+1)
        chunks[r].close()
        assert chunks[r].closed
    assert chunks[4].read() == '8'
    chunks[4].close()
    f.close()

def test_f_chunk_seek_tell():
    f = NamedTemporaryFile()
    f.write('01234567')
    c = utils.f_chunk(f, 4)[1]
    assert c.tell() == 0
    c.seek(0, 2)
    assert c.tell() == 4
    c.seek(0)
    assert c.tell() == 0
    c.seek(2, 1)
    assert c.tell() == 2
    f.close()



from fnmatch import fnmatch
import hashlib
import os
import re

class Matcher(object):
    """Matches strings based on patterns

    Internally the matcher object has a set of glob style patterns to match
    and to ignore, and a set of regexes to match and to ignore. Each of
    these are tested against the string 's' in the matches method. Ignore
    has the highest precedence. If both patterns and regexes are empty, the
    object assumes that the match is True (though the ignore sets can still
    overpower this assumption).

    """
    def __init__(self, patterns=None, ignore_patterns=None, regexes=None,
                 ignore_regexes=None):
        self.patterns = set(patterns or set())
        self.ignore_patterns = set(ignore_patterns or set())
        self.regexes = set(regexes or set())
        self.ignore_regexes = set(ignore_regexes or set())

    def matches(self, s):
        """Return whether this matcher matches string s"""
        # If neither patterns nor regexes is set, match everything
        matched = not self.patterns and not self.regexes
        for pattern in self.patterns:
            if fnmatch(s, pattern):
                matched = True
                break
        if not matched:
            for regex in self.regexes:
                if re.search(regex, s):
                    matched = True
                    break
        if matched:
            for pattern in self.ignore_patterns:
                if fnmatch(s, pattern):
                    return False
        if matched:
            for regex in self.ignore_regexes:
                if re.search(regex, s):
                    return False
        return matched

    def __add__(self, other):
        """Allow combining of matchers."""
        patterns = self.patterns | other.patterns
        ignore_patterns = self.ignore_patterns | other.ignore_patterns
        regexes = self.regexes | other.regexes
        ignore_regexes = self.ignore_regexes | other.ignore_regexes

        return Matcher(patterns, ignore_patterns, regexes, ignore_regexes)

    def __iadd__(self, other):
        return self.__add__(other)

def f_decorator(func):
    """Makes sure decorated function doesn't mess with file position."""
    def inner(f, *args, **kwargs):
        initial_pos = f.tell() # Get initial pos
        f.seek(0) # Rewind to beginning
        ret = func(f, *args, **kwargs) # Run decorated func
        f.seek(initial_pos) # Seek back to where it was
        return ret
    return inner

@f_decorator
def f_md5(f):
    """Return md5 hash of file like object."""
    m = hashlib.md5()
    while True:
        buf=f.read(8192)
        if not buf: break
        m.update(buf)
    return m.digest()

@f_decorator
def f_sizeof(f):
    """Return size of file like object."""
    f.seek(0, 2)
    return f.tell()

# Note: f *must* be a real file, not just file-like
# as f_chunk re-opens it using f.name
def f_chunk(f, chunk_size):
    """Return file like object split into chunk-sized file like objects"""
    class FChunk(object):
        """Mimics file interface for subset of a real file."""
        def __init__(self, f, start, size):
            self._f = open(f.name, 'rb')
            self._f.seek(start, 0)
            self._start = start
            self._end = start+size
            self._size = size

        def read(self, size=0):
            remaining = self._end - self._f.tell()
            if size <= 0 or size > remaining:
                return self._f.read(remaining)
            else:
                return self._f.read(size)

        def seek(self, offset, whence=0):
            if whence == 0:
                self._f.seek(self._start+offset, 0)
            elif whence == 1:
                self._f.seek(offset, 1)
            elif whence == 2:
                self._f.seek(self._end+offset, 0)
            else:
                raise IOError

        def tell(self):
            return self._f.tell() - self._start

        @property
        def closed(self):
            return self._f.closed

        def close(self):
            self._f.close()

    full_size = f_sizeof(f)
    num_chunks = (full_size+chunk_size-1)/chunk_size # Round up div

    chunks = []
    start = 0
    for r in range(num_chunks):
        # If last block is smaller than chunk_size
        if full_size < (start + chunk_size):
            chunk = FChunk(f, start, full_size-start)
        else:
            chunk = FChunk(f, start, chunk_size)
        chunks.append(chunk)
        start += chunk_size
    return chunks

def os_walk_relative(src):
    """Return list of all file paths in src relative to src."""
    for root, dirs, files in os.walk(src):
        for f in files:
            full_path = os.path.join(root, f)
            yield os.path.relpath(full_path, src)


import argparse
import logging
import textwrap
import sys
import os

from s3tup.parse import load_config, parse_config

log = logging.getLogger('s3tup')
title = (
    "\n"
    "        _____ __             \n"
    "   ____|__  // /___  ______  \n"
    "  / ___//_ </ __/ / / / __ \ \n"
    " /__  /__/ / /_/ /_/ / /_/ / \n"
    "/____/____/\__/\____/ /___/  \n"
    "                   /_/       \n"
)

parser = argparse.ArgumentParser(description=
    's3tup: Declarative configuration management and deployment tool for'
    ' Amazon S3')
parser.add_argument(
    'config',
    help='path to your configuration file')
parser.add_argument(
    '--dryrun',
    action='store_true',
    help='run s3tup without actually... running s3tup')
parser.add_argument(
    '--rsync',
    action='store_true',
    help='only upload/delete keys that have been modified/removed, don\'t'
         ' sync|redirect keys and don\'t sync bucket attributes')
parser.add_argument(
    '-c',
    metavar='CONCURRENCY',
    type=int,
    help='concurrency')
verbosity = parser.add_mutually_exclusive_group()
verbosity.add_argument(
    '-v', '--verbose',
    action='store_true',
    help='increase output verbosity')
verbosity.add_argument(
    '-q', '--quiet',
    action='store_true',
    help='silence all output')
parser.add_argument(
    '--access_key_id',
    help='your aws access key id. unnescessary if your AWS_ACCESS_KEY_ID env var is set.')
parser.add_argument(
    '--secret_access_key',
    help='your aws secret access key. unnescessary if your AWS_SECRET_ACCESS_KEY env var is set.')

def main():
    args = parser.parse_args()

    if not (args.quiet or args.verbose):
        handler = logging.StreamHandler()
        handler.setFormatter(WrappedFormatter('%(message)s'))
        log.addHandler(handler)
        log.setLevel(logging.INFO)
    elif args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s',
                            level=logging.DEBUG)

    try:
        run(args.config, args.dryrun, args.rsync, args.c,
            args.access_key_id, args.secret_access_key)
    except Exception as e:
        if args.verbose:
            raise
        log.error('{}: {}'.format(sys.exc_info()[0].__name__, e))
        sys.exit(1)

def run(config, dryrun=False, rsync=False, concurrency=None,
        access_key_id=None, secret_access_key=None,):

    if access_key_id is not None:
        os.environ['AWS_ACCESS_KEY_ID'] = access_key_id
    if secret_access_key is not None:
        os.environ['AWS_SECRET_ACCESS_KEY'] = secret_access_key

    config = load_config(config)
    buckets = parse_config(config)

    log.info(title)

    for b in buckets:
        if concurrency is not None:
            b.conn.concurrency = concurrency
        b.sync(dryrun=dryrun, rsync=rsync)

class WrappedFormatter(logging.Formatter):
    """Wraps log lines at 78 chars."""
    def format(self, record):
        formatted = super(WrappedFormatter, self).format(record)
        split = formatted.split('\n')
        return '\n'.join([textwrap.fill(l, 78) for l in split])


from binascii import hexlify
import hashlib
import logging
import os

from s3tup.exception import ActionConflict
import s3tup.utils as utils

log = logging.getLogger('s3tup.rsync')

class ActionPlan(object):
    """ Stores a set of actions to be executed on keys.

    Buckets know how to execute action plans, and action
    plans know when there are conflicting actions on keys.
    """

    def __init__(self):
        self._actions = {}

    # Ugly... but works and passes tests
    def _add_action(self, action_type, key, *args):
        new_val = [action_type,]+list(args)
        old_val = self._actions.pop(key, None)

        if old_val == None:
            self._actions[key] = new_val
        elif old_val[0] == 'delete':
            if action_type != 'delete':
                # don't complain if delete is overwritten
                pass
            self._actions[key] = new_val
        else:
            if action_type == 'delete':
                self._actions[key] = old_val
            else:
                if new_val == old_val:
                    self._actions[key] = old_val
                else:
                    msg = "Conflicting actions set for key '{}':\n".format(key)
                    for v in (new_val, old_val):
                        if v[0] in ("delete", "sync"):
                            msg += '{}, '.format(v[0])
                        else:
                            msg += '{} -> {}, '.format(v[0], v[1])
                    msg = msg[:-2]
                    raise ActionConflict(msg)

    def remove_actions(self, *action_types):
        toremove = []
        for key,v in self._actions.items():
            if v[0] in action_types:
                toremove.append(key)
        for k in toremove:
            self._actions.pop(k) #

    def delete(self, key):
        self._add_action('delete', key)

    def sync(self, key):
        self._add_action('sync', key)

    def redirect(self, key, url):
        self._add_action('redirect', key, url)

    def upload(self, key, path):
        self._add_action('upload', key, path)

    @property
    def affected_keys(self):
        return self._actions.keys()

    @property
    def to_upload(self):
        for k,v in self._actions.items():
            if v[0] == 'upload':
                yield k, v[1]

    @property
    def to_redirect(self):
        for k,v in self._actions.items():
            if v[0] == 'redirect':
                yield k, v[1]

    @property
    def to_delete(self):
        for k,v in self._actions.items():
            if v[0] == 'delete':
                yield k

    @property
    def to_sync(self):
        for k,v in self._actions.items():
            if v[0] == 'sync':
                yield k

    def __add__(self, other):
        new = ActionPlan()
        for old in (other, self):
            for key,v in old._actions.items():
                if v[0] == 'redirect':
                    new.redirect(key, v[1])
                if v[0] == 'upload':
                    new.upload(key, v[1])
                if v[0] == 'delete':
                    new.delete(key)
                if v[0] == 'sync':
                    new.sync(key)
        return new

    def __iadd__(self, other):
        return self.__add__(other)


class RsyncPlanner(object):
    """Container for RsyncConfigs."""

    def __init__(self, rsync_configs=None):
        self.configs = rsync_configs or []

    def plan(self, remote_keys):
        plan = ActionPlan()
        for config in self.configs:
            plan += config.plan(remote_keys)
        return plan

class RsyncConfig(object):

    def __init__(self, src=None, dest=None, delete=False, matcher=None):
        self.src = src
        self.dest = dest
        self.delete = delete
        self.matcher = matcher or utils.Matcher()

    def plan(self, remote_keys):
        remote_key_names = set(remote_keys.keys())
        new = set()
        modified = set()
        unmodified = set()
        for k in self._get_local_key_names():
            if k not in remote_key_names:
                new.add(k)
            else:
                if (self._md5_cmp(remote_keys[k].md5, k)):
                    unmodified.add(k)
                else:
                    modified.add(k)
        removed = remote_key_names - (modified | unmodified)

        plan = ActionPlan()
        for k in new | modified:
            plan.upload(k, self._get_local_path_from_key(k))
        for k in unmodified:
            plan.sync(k)
        if self.delete:
            for k in removed:
                plan.delete(k)
        return plan

    def _md5_cmp(self, s3_md5, key):
        local_path = self._get_local_path_from_key(key)
        with open(local_path, 'rb') as f:
            if not '-' in s3_md5:
                local_md5 = hexlify(utils.f_md5(f))
            else:
                chunks = utils.f_chunk(f, 5242880)
                m = hashlib.md5()
                for chunk in chunks:
                    m.update(utils.f_md5(chunk))
                local_md5 = "{}-{}".format(hexlify(m.digest()), len(chunks))
            return local_md5 == s3_md5        

    def _get_local_key_names(self):
        src = self.src or '.'
        dest = self.dest or '.'
        for path in utils.os_walk_relative(src):
            if not self.matcher.matches(path):
                continue
            yield os.path.normpath(os.path.join(dest, path))

    def _get_local_path_from_key(self, key):
        src = self.src or '.'
        dest = self.dest or '.'
        return os.path.normpath(os.path.join(src, os.path.relpath(key, dest)))


from nose.tools import raises

from s3tup.bucket import Bucket

from utils import ConnMock

@raises(TypeError)
def test_bucket_init_invalid_kwarg():
    Bucket(None, None, None, invalid='invalid')

@raises(TypeError)
def test_bucket_init_missing_name():
    Bucket(None)

def test_bucket_init_success():
    b = Bucket(None, 'test', acl='test')
    assert b.name == 'test'
    assert b.acl == 'test'


from nose.tools import raises

from s3tup.rsync import ActionPlan, RsyncConfig
from s3tup.exception import ActionConflict

# ACTION PLAN

def test_action_plan_delete():
    ap = ActionPlan()
    ap.delete('test1')
    ap.delete('test2')
    ap.delete('test1')
    r = list(ap.to_delete)
    assert len(r) == 2
    assert 'test1' in r
    assert 'test2' in r

def test_action_plan_delete_after_upload():
    ap = ActionPlan()
    ap.upload('test', 'path')
    ap.delete('test')
    assert len(list(ap.affected_keys)) == 1
    assert len(list(ap.to_delete)) == 0
    assert len(list(ap.to_upload)) == 1

def test_action_plan_upload_after_delete():
    ap = ActionPlan()
    ap.delete('test')
    ap.upload('test', 'path')
    assert len(list(ap.affected_keys)) == 1
    assert len(list(ap.to_delete)) == 0
    assert len(list(ap.to_upload)) == 1

def test_action_plan_upload_after_upload_same():
    ap = ActionPlan()
    ap.upload('test', 'path')
    ap.upload('test', 'path')
    assert len(list(ap.affected_keys)) == 1
    assert len(list(ap.to_upload)) == 1

@raises(ActionConflict)
def test_action_plan_upload_after_upload_different():
    ap = ActionPlan()
    ap.upload('test', 'path')
    ap.upload('test', 'different_path')

def test_action_plan_merge():
    ap1 = ActionPlan()
    ap2 = ActionPlan()
    ap1.delete('test1')
    ap2.upload('test2', 'path')
    ap3 = ap1 + ap2
    assert len(list(ap3.affected_keys)) == 2
    assert len(list(ap3.to_upload)) == 1
    assert len(list(ap3.to_delete)) == 1

@raises(ActionConflict)
def test_action_plan_merge_conflict():
    ap1 = ActionPlan()
    ap2 = ActionPlan()
    ap1.redirect('test1', 'url')
    ap2.upload('test1', 'path')
    ap1 + ap2

# RSYNC CONFIG

def test_rsync_config_get_local_path_from_key():
    r = RsyncConfig()
    assert r._get_local_path_from_key('key') == 'key'
    r.src = 'src'
    assert r._get_local_path_from_key('key') == 'src/key'
    r.dest = 'dest'
    assert r._get_local_path_from_key('dest/key') == 'src/key'
    r.src = None
    assert r._get_local_path_from_key('dest/key') == 'key'


from nose.tools import raises

from s3tup import parse
from s3tup.exception import ConfigParseError
from s3tup.rsync import RsyncConfig

# parse_config

@raises(ConfigParseError)
def test_parse_config_non_list_config():
    parse.parse_config('invalid')

def test_parse_config_success():
    bucket_config = {'bucket': 'test'}
    config = [bucket_config for r in range(10)]
    buckets = parse.parse_config(config)
    assert len(buckets) == 10
    assert buckets[0].name == 'test'

# parse_bucket

@raises(ConfigParseError)
def test_parse_bucket_invalid_kwarg():
    b = {'bucket': 'test', 'invalid': 'kwarg'}
    parse.parse_bucket(b)

@raises(ConfigParseError)
def test_parse_bucket_no_name_provided():
    parse.parse_bucket({})

def test_parse_bucket_success():
    b = {'bucket': 'test', 'region': 'test'}
    bucket = parse.parse_bucket(b)
    assert bucket.name == 'test'
    assert bucket.region == 'test'

# parse_rsync

def test_parse_rsync_string_input():
    r = parse.parse_rsync('test')
    assert len(r.configs) == 1
    assert isinstance(r.configs[0], RsyncConfig)
    assert r.configs[0].src == 'test'

def test_parse_rsync_dict_input():
    d = {'src': 'test', 'delete': True}
    r = parse.parse_rsync(d)
    assert len(r.configs) == 1
    assert isinstance(r.configs[0], RsyncConfig)
    assert r.configs[0].src == 'test'
    assert r.configs[0].delete

def test_parse_rsync_list_input():
    d = {'src': 'test'}
    l = [d for r in range(5)]
    r = parse.parse_rsync(l)
    assert len(r.configs) == 5
    assert isinstance(r.configs[0], RsyncConfig)
    assert r.configs[0].src == 'test'
    assert not r.configs[0].delete

# parse_rsync_object

@raises(ConfigParseError)
def test_parse_rsync_object_invalid_kwarg():
    d = {'src': 'test', 'invalid': True}
    parse.parse_rsync_object(d)

def test_parse_rsync_object_success():
    d = {'src': 'test', 'delete': True, 'patterns': ['test',]}
    rs = parse.parse_rsync_object(d)
    assert rs.src == 'test'
    assert rs.delete
    assert 'test' in rs.matcher.patterns

# parse_key_config

@raises(ConfigParseError)
def test_non_list_config():
    parse.parse_key_config({'invalid': 'invalid'})

def test_success():
    c = {'patterns': ['*.py'], 'reduced_redundancy': True}
    key_config = [c.copy() for r in range(10)]
    factory = parse.parse_key_config(key_config)
    k = factory.make_key(None, None, 'test')
    assert not k.reduced_redundancy
    k = factory.make_key(None, None, 'test.py')
    assert k.reduced_redundancy

# parse_key_configurator

@raises(ConfigParseError)
def test_parse_key_configurator_invalid_kwarg():
    parse.parse_key_configurator({'invalid': 'invalid'})

def test_parse_key_configurator_success():
    c = {'patterns': ['*.py'], 'reduced_redundancy': True}
    configurator = parse.parse_key_configurator(c)
    assert configurator.effects_key('test.py')
    assert not configurator.effects_key('test.md')
    assert configurator.reduced_redundancy

# extract_matcher

def test_extract_matcher_all_none():
    matcher = parse.extract_matcher({})[0]
    assert matcher == None

def test_extract_matcher_popped_from_config():
    c = {'patterns': [], 'ignore_patterns': [], 'regexes': [],
         'ignore_regexes': [], 'test': 'test'}
    config = parse.extract_matcher(c)[1]
    assert len(config) == 1
    assert config['test'] == 'test'

def test_extract_matcher_success():
    matcher = parse.extract_matcher({'patterns': ['test'],})[0]
    assert len(matcher.patterns) == 1
    assert 'test' in matcher.patterns


from StringIO import StringIO

from nose.tools import raises

from s3tup.key import Key, KeyConfigurator
from s3tup.utils import Matcher

from utils import ConnMock

# KEY CONFIGURATOR

def test_key_configurator_configure_key():
    c = KeyConfigurator()
    c.reduced_redundancy = True
    c.encrypted = False
    c.cache_control = 'test'
    c.canned_acl = 'public-read'
    c.metadata = {'some':'metadata'}

    # Create new key and run configurator on it
    k = Key(None, 'bucket', 'key', canned_acl='private', encrypted=True,
            metadata={'some':'otherdata'}, cache_control='original')
    k = c.configure_key(k)

    # Confirm configurator worked as expected
    assert k.reduced_redundancy
    assert not k.encrypted
    assert k.cache_control == 'test'
    assert k.canned_acl == 'public-read'
    assert k.metadata['some'] == 'metadata'

def test_key_configurator_effects_key():
    c = KeyConfigurator(matcher=Matcher(['*.py']))
    assert c.effects_key('test.py')
    assert not c.effects_key('test.md')

# KEY

def test_key_pretty_path():
    k = Key(None, 'bucket', 'key')
    assert k.pretty_path == 's3://bucket/key'

def test_key_delete():
    conn = ConnMock()
    key = Key(conn, 'test', 'test')
    key.delete()
    conn.make_request.assert_called_once_with(
        'DELETE',
        'test',
        'test'
    )

def test_key_redirect():
    conn = ConnMock()
    key = Key(conn, 'test', 'test')
    key.redirect('url')
    conn.make_request.assert_called_once_with(
        'PUT',
        'test',
        'test',
        headers={'x-amz-website-redirect-location':'url'}
    )

def test_key_sync():
    conn = ConnMock()
    key = Key(conn, 'test', 'test')
    key.encrypted = True
    key.sync()
    expected_headers = {
        'x-amz-copy-source':'/test/test',
        'x-amz-metadata-directive':'REPLACE',
        'x-amz-server-side-encryption':'AES256'
    }
    conn.make_request.assert_called_once_with(
        'PUT',
        'test',
        'test',
        headers=expected_headers,
    )

def test_key_upload():
    conn = ConnMock()
    key = Key(conn, 'test', 'test')
    s = StringIO('test')
    key.upload(s)
    conn.make_request.assert_called_once_with(
        'PUT',
        'test',
        'test',
        data='test'
    )

def test_key_init():
    k = Key(
        None,
        'key',
        'bucket',
        cache_control = 'public',
        canned_acl = 'public-read',
        content_disposition = 'attachment',
        content_encoding = 'gzip',
        content_type = 'text/html',
        content_language = 'mi, en',
        encrypted = True,
        expires = 'Thu, 01 Dec 1994 16:00:00 GMT',
        metadata = {'k1': 'v1', 'k2': 'v2'},
        reduced_redundancy = True,
    )
    assert k.cache_control == 'public'
    assert k.content_encoding == 'gzip'
    assert k.content_disposition == 'attachment'
    assert k.content_language == 'mi, en'
    assert k.content_type == 'text/html'
    assert k.expires == 'Thu, 01 Dec 1994 16:00:00 GMT'
    assert k.canned_acl == 'public-read'
    assert k.encrypted == True
    assert k.reduced_redundancy == True
    assert k.metadata['k1'] == 'v1'
    assert k.metadata['k2'] == 'v2'

@raises(TypeError)
def test_key_init_invalid_kwarg():
    Key(None, None, None, invalid_kwarg=True)

def test_key_init_defaults():
    k = Key(None, None, None)
    assert not k.encrypted
    assert not k.reduced_redundancy
    assert k.metadata == {}

def test_key_headers():
    k = Key(None, 'bucket', 'key')

    k.cache_control = 'public'
    k.canned_acl = 'public-read'
    k.content_disposition = 'attachment'
    k.content_encoding = 'gzip'
    k.content_type = 'text/html'
    k.content_language = 'mi, en'
    k.encrypted = True
    k.expires = 'Thu, 01 Dec 1994 16:00:00 GMT'
    k.metadata = {'k1': 'v1', 'k2': 'v2'}
    k.reduced_redundancy = True

    headers = k.get_headers()

    assert headers['cache-control'] == 'public'
    assert headers['content-encoding'] == 'gzip'
    assert headers['content-disposition'] == 'attachment'
    assert headers['content-language'] == 'mi, en'
    assert headers['content-type'] == 'text/html'
    assert headers['expires'] == 'Thu, 01 Dec 1994 16:00:00 GMT'
    assert headers['x-amz-acl'] == 'public-read'
    assert headers['x-amz-server-side-encryption'] == 'AES256'
    assert headers['x-amz-storage-class'] == 'REDUCED_REDUNDANCY'
    assert headers['x-amz-meta-k1'] == 'v1'
    assert headers['x-amz-meta-k2'] == 'v2'
    assert len(headers) == 11

def test_key_headers_acl_is_none():
    k = Key(None, 'bucket', 'key')
    k.acl = None

    headers = k.get_headers()
    assert headers['x-amz-acl'] == 'private'

def test_key_headers_mimetype_guessing():
    k = Key(None, 'bucket', 'test.css')
    headers = k.get_headers()
    assert headers['content-type'] == 'text/css'




from setuptools import setup

setup(
    name='s3tup',
    version='0.1.0',
    description='Declarative configuration management and deployment tool for amazon s3',
    keywords='s3 declarative config deploy',
    url='http://github.com/HeyImAlex/s3tup',
    author='Alex Guerra',
    author_email='alex@heyimalex.com',
    license='MIT',
    classifiers=[
      'Development Status :: 3 - Alpha',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 2.7',
    ],
    packages=['s3tup'],
    install_requires=[
        'requests',
        'argparse',
        'pyyaml',
        'beautifulsoup4',
        'gevent>=1.0',
    ],
    entry_points = {
      'console_scripts': ['s3tup=s3tup.cli:main'],
    },
    zip_safe=False,
    test_suite = 'nose.collector',
    tests_require = ['nose>=1.0', 'mock>=1.0']
)


from collections import namedtuple
import logging

from bs4 import BeautifulSoup

from s3tup.key import KeyFactory, redirect_key, delete_key
from s3tup.rsync import ActionPlan
import s3tup.constants as constants

log = logging.getLogger('s3tup.bucket')

class Bucket(object):
    """
    Encapsulates configuration for an s3 bucket and its keys. It contains a
    KeyFactory, which handles key configuration, an RsyncPlanner, which
    decides which actions to take on keys, and attributes (listed
    in constants.BUCKET_ATTRS) that you can set, delete, modify, and then
    sync to s3 using the various sync methods provided.

    """
    def __init__(self, conn, name, key_factory=None,
                 rsync_planner=None, **kwargs):
        self.conn = conn
        self.name = name
        self.key_factory = key_factory or KeyFactory()
        self.rsync_planner = rsync_planner
        self.redirects = kwargs.pop('redirects', [])

        # Add all kwargs passed in that are named in 
        # constants.BUCKET_ATTRS to this object instance
        for k,v in kwargs.items():
            if k in constants.BUCKET_ATTRS:
                self.__dict__[k] = v
            else:
                msg = ("__init__() got an unexpected keyword"
                       " argument '{}'".format(k))
                raise TypeError(msg)

    def make_request(self, method, params=None, data=None, headers=None):
        """Convenience method for self.conn.make_request; has the bucket and
        key fields already filled in."""
        return self.conn.make_request(
            method=method,
            bucket=self.name,
            key=None,
            params=params,
            data=data,
            headers=headers
        )

    # KEY METHODS

    def make_key(self, key_name):
        return self.key_factory.make_key(self.conn, self.name, key_name)

    def sync_key(self, key_name):
        key = self.make_key(key_name)
        key.sync()

    def upload_key(self, key_name, f):
        key = self.make_key(key_name)
        key.upload(f)

    def redirect_key(self, key_name, redirect_url):
        redirect_key(self.conn, self.name, key_name, redirect_url)

    def delete_key(self, key_name):
        delete_key(self.conn, self.name, key_name)

    def delete_keys(self, key_names):
        """Delete a list of keys from this bucket.

        key_names is a list of str key names to be deleted. S3's delete
        operation has a limit of 1000 keys per request so this method
        handles paging as well.

        """
        for i in range(0, len(key_names), 1000):
            data = ('<?xml version="1.0" encoding="UTF-8"?>\n'
                    '<Delete>\n  <Quiet>true</Quiet>\n')
            for k in key_names[i:i+1000]:
                log.info('delete: s3://{}/{}'.format(self.name, k))
                data += '  <Object><Key>{}</Key></Object>\n'.format(k)
            data += '</Delete>'
            # TODO: Make this joinable.
            self.make_request('POST', 'delete', data=data)

    # Named get_remote_keys instead of just overriding __iter__
    # to avoid ambiguity. Returns dicts instead of Key objects for the
    # same reason. Requests can't be parallel as each depends on
    # the marker from the last.
    def get_remote_keys(self, prefix=None):
        """Generate list of dicts representing all keys in this s3 bucket.

        Each dict returned contains fields 'name', 'md5', 'size', and
        'modified'. Paging is handled automatically. Optional (str)
        prefix param will limit the results to those keys prefixed by it.

        """
        KeyTuple = namedtuple('KeyTuple', ['name', 'md5', 'size', 'modified'])
        keys = {}
        more = True
        marker = None
        while more is True:
            params = {'marker': marker, 'prefix': prefix}
            r = self.conn.make_request('GET', self.name, params=params)
            root = BeautifulSoup(r.text).find('listbucketresult')
            for c in root.find_all('contents'):
                key = c.find('key').text
                marker = key
                modified = c.find('lastmodified').text
                size = int(c.find('size').text)
                md5 = c.find('etag').text.replace('"', '')
                keys[key] = KeyTuple(key, md5, size, modified)
            more = root.find('istruncated').text == 'true'
        return keys

    # SYNC METHODS

    def sync(self, dryrun=False, rsync=False):
        """Sync everything.

        Takes every applicable attribute set on this Bucket object and
        configures its respective s3 bucket (defined by self.name) to match
        it. Optional rsync only mode will only run rsync (no setting bucket
        configuration, no syncing unmodified keys, no making redirects).

        """
        log.info("syncing bucket '{}'...".format(self.name))

        self.create()
        if not rsync:
            self.sync_bucket(dryrun=dryrun)
        self.sync_keys(dryrun=dryrun, rsync=rsync)

        log.info("bucket '{}' sucessfully synced!\n".format(self.name))

    def create(self):
        try:
            headers = {'x-amz-acl': self.canned_acl}
        except AttributeError:
            headers = None
        try:
            data = ('<CreateBucketConfiguration '
                    'xmlns="http://s3.amazonaws.com/doc/2006-03-01/">\n'
                    '  <LocationConstraint>{}</LocationConstraint>\n'
                    '</CreateBucketConfiguration>').format(self.region)
        except AttributeError:
            data = None
        self.make_request('PUT', headers=headers, data=data)

    def sync_bucket(self, dryrun=False):
        if dryrun:
            tmp = self.make_request
            self.make_request = lambda *args, **kwargs:None
        self.conn.join([
            self.sync_acl,
            self.sync_cors,
            self.sync_lifecycle,
            self.sync_logging,
            self.sync_notification,
            self.sync_policy,
            self.sync_tagging,
            self.sync_versioning,
            self.sync_website,
        ])
        if dryrun:
            self.make_request = tmp

    def sync_keys(self, dryrun=False, rsync=False):
        plan = self._make_action_plan(rsync)
        if not dryrun:
            self._execute_action_plan(plan)
        else:
            for k, path in plan.to_upload:
                log.info("upload: {} <- {}".format(k, path))
            for k in plan.to_sync:
                log.info("sync: {}".format(k))
            for k, url in plan.to_redirect:
                log.info("redirect: {} -> {}".format(k, url))
            for k in plan.to_delete:
                log.info("delete: {}".format(k))

    def _make_action_plan(self, rsync=False):
        remote_keys = self.get_remote_keys()

        if self.rsync_planner:
            plan = self.rsync_planner.plan(remote_keys)
        else:
            plan = ActionPlan()

        # Add in redirects
        for key, url in self.redirects:
            plan.redirect(key, url)

        # Sync all keys with no action yet associated
        affected_keys = set(plan.affected_keys)
        old_keys = set(remote_keys.keys())
        for k in (old_keys-affected_keys):
            plan.sync(k)

        if rsync:
            plan.remove_actions('sync', 'redirect')

        return plan

    def _execute_action_plan(self, plan):
        actions = []
        for k, path in plan.to_upload:
            actions.append([self.upload_key, k, open(path, 'rb')])
        for k in plan.to_sync:
            actions.append([self.sync_key, k])
        for k, url in plan.to_redirect:
            actions.append([self.redirect_key, k, url])
        actions.append([self.delete_keys, list(plan.to_delete)])
        self.conn.join(actions)

    # INDIVIDUAL BUCKET SYNCING METHODS
    #
    # Each checks if this bucket has its respective attr set and, if it does,
    # proceeds to sync that value with the s3 bucket. Each will return s3's
    # response in the form of a requests.Response object if the attr is set,
    # and if not will return False. These map directly to the fields
    # defined in the bucket config section of the readme, so if you're
    # looking for details on what values are allowed check there.

    def sync_acl(self):
        try: acl = self.acl
        except AttributeError: return False
        
        if acl is not None:
            log.info("set xml acl")
            return self.make_request('PUT', 'acl', data=acl)
        else:
            log.info("revert to default acl")
            return self.make_request('PUT', 'acl',
                                     headers={"x-amz-acl":"private"})

    def sync_cors(self):
        try: cors = self.cors
        except AttributeError: return False

        if cors is not None:
            log.info("set cors configuration")
            return self.make_request('PUT', 'cors', data=cors)
        else:
            log.info("delete cors configuration")
            return self.make_request('DELETE', 'cors')

    def sync_lifecycle(self):
        try: lifecycle = self.lifecycle
        except AttributeError: return False

        if lifecycle is not None:
            log.info("set lifecycle configuration")
            return self.make_request('PUT', 'lifecycle', data=lifecycle)
        else:
            log.info("delete lifecycle configuration")
            return self.make_request('DELETE', 'lifecycle')

    def sync_logging(self):
        try: logging = self.logging
        except AttributeError: return False

        if logging is not None:
            log.info("set logging configuration")
            data = logging
        else:
            log.info("delete logging configuration")
            data = ('<?xml version="1.0" encoding="UTF-8"?>\n'
                    '<BucketLoggingStatus '
                    'xmlns="http://doc.s3.amazonaws.com/2006-03-01"/>')
        return self.make_request('PUT', 'logging', data=data)

    def sync_notification(self):
        try: notification = self.notification
        except AttributeError: return False

        if notification is not None:
            log.info("set notification configuration")
            data = notification
        else:
            log.info("delete notification configuration")
            data = '<NotificationConfiguration />'
        return self.make_request('PUT', 'notification', data=data)

    def sync_policy(self):
        try: policy = self.policy
        except AttributeError: return False

        if policy is not None:
            log.info("set bucket policy")
            return self.make_request('PUT', 'policy', data=policy)
        else:
            log.info("delete bucket policy")
            return self.make_request('DELETE', 'policy')

    def sync_tagging(self):
        try: tagging = self.tagging
        except AttributeError: return False

        if tagging is not None:
            log.info("set bucket tags")
            return self.make_request('PUT', 'tagging', data=tagging)
        else:
            log.info("delete bucket tags")
            return self.make_request('DELETE', 'tagging')

    def sync_versioning(self):
        try: versioning = self.versioning
        except AttributeError: return False

        if versioning:
            log.info("enable versioning")
            status = 'Enabled'
        else:
            log.info("suspend versioning")
            status = 'Suspended'
        data = ('<VersioningConfiguration '
                'xmlns="http://s3.amazonaws.com/doc/2006-03-01/">\n'
                '  <Status>{}</Status>\n'
                '</VersioningConfiguration>').format(status)
        return self.make_request('PUT', 'versioning', data=data)

    def sync_website(self):
        try: website = self.website
        except AttributeError: return False

        if website is not None:
            log.info("set website configuration")
            return self.make_request('PUT', 'website', data=website)
        else:
            log.info("delete website configuration")
            return self.make_request('DELETE', 'website')